


################################################################################################
#
#  Uses Batch Gradient Descent to estimate coefficients of a multiple linear regression model.   
#  Uses only the mathematics dataset related to student performance in a school in Portugal [1].
#
#  The models aims at only estimating the parameters to examine the relationship between the final 
#  grade of students in a class in secondary education (dependent variable) and the number of 
#  times students failed the class and their days of absence of the class. Estimates the 
#  parameters of the linear model:
#  G3 = b1*failures + b2*absences + b0
#
#  Stopping criterion is the number of iterations.
# 
#  Does not do any evaluation with respect to its predictive or prescriptive goals. 
#  The only evaluation done is a comparison of the estimated coefficients with the 
#  coefficients estimated with the OLS method. This is done to see the impact of alpha and number
#  of iterations.
#
#  References: 
#     1) Student Performance (math dataset), https://archive.ics.uci.edu/dataset/320/student+performance
# 
#
#  Required libraries:
#     1) corrplot
#     2) car
# 
#
#  v0.3/mmt/Nov 2024
#
################################################################################################


# Close any graphic device left open
graphics.off()

# Cleanup the environment
# all=TRUE is optional
rm(list = ls())

# Clear console
cat("\014")

# Note: Change this iff needed.
setwd("C:/Users/user/Desktop")







# A visual exploratory tool on correlation matrix
library(corrplot)
# For vif()
library(car)



DEFAULT_GRAPHICAL_PARAMS <- par(no.readonly = TRUE)
STUDENT_DATA_FILE = "student-mat.csv"

# Batch gradient descent parameters
ALPHA = 0.00199 # Alpha value: the learning rate

# Number of iterations to carry out 
N_ITERATIONS = 3500 







##########################################################################################
# Function definitions
##########################################################################################


#' Calculates the value of the cost function J(θ) based on the current values
#' of thetas. 
#'
#' @details  Uses the formula
#'           J(θ) = SUM((indVars*thetas - depVar)^2) / 2n
#'           where n the number of observations/rows.
#'                            
#'
#' @param indVars a n x (k+1) MATRIX with the values of the independent 
#'               variables. k is the number of variables in the regression
#'               model. 
#' @param depVar a VECTOR with the n values of the dependent variable (one for each row). 
#' @param thetas a (k+1) x 1 MATRIX with the current values of the thetas (coefficients). 
#'              
#' @return A number representing the value of the cost function J(thetas)
#'          
CalculateCost <- function(indVars, depVar, thetas){
  # Πλήθος παρατηρήσεων
  m <- length(depVar)
  return( sum((indVars%*%thetas - depVar)^2) / (2*m) )
} # CalculateCost






#' Implements the Batch Gradient Descent algorithm. Stopping criterion is 
#' the number of iterations. 
#'
#' @details  Implements the Batch Gradient Descent algorithm from scratch. Stores
#'           the values of the cost function in a vector in order to 
#'           
#'           
#'                            
#'
#' @param indVars: a n x (k+1) MATRIX with the values of the independent 
#'                 variables. k is the number of independent variables used 
#'                 in the regression model. n is the number of observations 
#'                 i.e. number of rows in dataset) If the intercept needs to be estimated, 
#'                 indVars should contain an additional column with all values 
#'                 equal to 1 (hence the +1 in k+1).
#' @param depVar:  a VECTOR with the n values of the dependent variable (n=number of rows in
#'                 the dataset). 
#' @param thetas:  a (k+1) x 1 MATRIX with the initial values of the thetas (coefficients).
#'                 If NULL given, a random set of intial thetas in the range (, 1) is used 
#' @param alpha:   the learning rate. Must be a positive number. 
#' @param nIter:   number of iterations to execute
#'              
#' @return A list containing the estimated coefficients, all the values of the cost history and the number
#'         of iterations carried out.     
#'          
BatchGradientDescent <- function(indVars, depVar, initTheta=NULL, alpha=0.01, nIter=90){
  
  if (is.null(initTheta)){
    thetas <- runif(n = ncol(indVars), min = 0, max = 1)
  }else{
    thetas <- initTheta
  }
  
  m <- length(depVar)
  costHistory <- rep(0, nIter)
  
  for(i in 1:nIter){
    thetas <- thetas - alpha*(1/m)*(t(indVars)%*%(indVars%*%thetas - depVar))
    costHistory[i]  <- CalculateCost(indVars, depVar, thetas)
    cat('Iteration ', i, ') thetas: ', thetas, '\n', sep=' ')
  } # for
  
  return(list("coefficients"=thetas, "costs"=costHistory, "niter"=nIter))
} # gradientDescent







##########################################################################################
# Execution starts from here
##########################################################################################


# Adding quot='"' argument because strings in csv file are between ""
studentMathematicsPerformanceData <- read.csv(STUDENT_DATA_FILE, sep=";", 
                                              header=TRUE, 
                                              quot='"', 
                                              stringsAsFactors = TRUE)



##########################################################################################
# Preprocessing
##########################################################################################

# Dependent variable
g3Grade <- studentMathematicsPerformanceData[, "G3"]

# Create a matrix with the independent variables. 
# Since we also require the estimation of the constant term, a column with all values 
# equal to 1 is added. 
# In the resulting matrix, first column is the contant term in the regression, second
# column are the failures and third column the absences.
# This is important to remember since this will allow us to see to which
# independent variable each coefficient belongs.
indVariables<- cbind( rep(1, nrow(studentMathematicsPerformanceData)), 
                      studentMathematicsPerformanceData[, c("failures", "absences")] ) 

##########################################################################################
# Linear model estimation
##########################################################################################



# Estimating coefficients using Batch Gradient Descent. 
bgdResult <- BatchGradientDescent(as.matrix(indVariables), g3Grade, NULL, ALPHA, N_ITERATIONS)


##########################################################################################
# Evaluation 
##########################################################################################

# Checking if learning rate is appropriate (Elbow method)
plot(bgdResult$costs, xlab="Επαναλήψεις", ylab="J(θ)" )  



# Comparing estimated coefficients with coefficients estimated using OLS.


# Estimating the linear regression model using the OLS method:
# G3 = b1*failures + b2*absences + b0 
# - G3: the student's final grade of the class (in the 0-20 range)
# - failures: number of times the student failed the class
# - absences: number of school absences 
#
# Using the same dataset for fitting the model.
studentPerfModel <-  lm(G3 ~ failures + absences, data=studentMathematicsPerformanceData)



# Display coefficients estimated by Batch Gradient Descent and OLS
# in a way to easier compare them: in a data frame with row labels. 
# Also shows pvalues as they are calculated by OLS.
comparisonDF <- data.frame(BGD=as.vector(bgdResult$coefficients), 
                           OLS=as.vector(coef(summary(studentPerfModel))[,1]), 
                           pvaluesOLS=as.vector(coef(summary(studentPerfModel))[,4]))

print(comparisonDF)




