


################################################################################################
#
#  Uses OLS to estimate coefficients of a multiple linear regression model.   
#  Uses only the mathematics dataset related to student performance in a school in Portugal [1].
#
#  The models aims at descriptive modelling, examining the relationship between the final 
#  grade of students in a class in secondary education (dependent variable) and the number of 
#  times students failed the class and their days of absence of the class. Estimates the 
#  linear model:
#  G3 = b1*failures + b2*absences + b0
#
#  References: 
#     1) Student Performance (math dataset), https://archive.ics.uci.edu/dataset/320/student+performance
# 
#  
# 
#
#  v0.3/mmt/May 2022
#
################################################################################################


# Close any graphic device left open
graphics.off()

# Cleanup the environment
# all=TRUE is optional
rm(list = ls())

# Clear console
cat("\014")



# A visual exploratory tool on correlation matrix
library(corrplot)
# For vif()
library(car)



setwd("C:/Users/user/Desktop")



DEFAULT_GRAPHICAL_PARAMS <- par(no.readonly = TRUE)
STUDENT_DATA_FILE = "student-mat.csv"






# Adding quot='"' argument because strings in csv file are between ""
studentMathematicsPerformanceData <- read.csv(STUDENT_DATA_FILE, sep=";", 
                                              header=TRUE, 
                                              quot='"', 
                                              stringsAsFactors = TRUE)



##########################################################################################
# Linear model estimation
##########################################################################################

# Estimating the linear regression model using the OLS method:
# G3 = b1*failures + b2*absences + b0 
# - G3: the student's final grade of the class (in the 0-20 range)
# - failures: number of times the student failed the class
# - absences: number of school absences 
#
# Using the entire dataset for fitting the model.
studentPerfModel <-  lm(G3 ~ failures + absences, data=studentMathematicsPerformanceData)


##########################################################################################
# Evaluation (Checking statistical significance)
##########################################################################################

# Print only the coefficients using the coef() function
# (it's an alternative way of gettig coefficients)
print(coef(studentPerfModel))

# Print complete results of linear regression.
# This will display also p-values of independent variables.
print(summary(studentPerfModel))

##########################################################################################
# Evaluation: Checking linear regression assumptions since we do descriptive modelling
##########################################################################################

##########################################################################################
#
# Assumption 1: Checking linearity.
#
# Answers the question whether the variables have really a linear relationship. They could
# have a non-linear relationship. 
#
# Linearity is checked by plotting a fitted vs real values scatter plot and a fitted values
# vs residuals plot. The first scatter plot should display a linear trend while the second
# plot should not reveal any pattern and be randomly distributed around 0.
# 
# If this assumption is violated, this is serious. The linear model is not justified and
# the estimated values cannot be considered valid.
# 
##########################################################################################

# Sets some graphical parameters.
# Splits figure into different sections and each subsequent plot will
# be displayed in a separate section. 
# Here layout is split into 1 row, 2 columns to display 2 plots side by side.
par(mfrow=c(1,2))

# Fitted values vs real FoodExpenditure values
plot(studentPerfModel$fitted.values,
     studentMathematicsPerformanceData$G3, xlab="Predicted values", ylab="Real values")

# Fitted values vs residuals
plot(studentPerfModel$fitted.values,
     studentPerfModel$residuals, xlab="Predicted values", ylab="Residuals")



##########################################################################################
#
# Assumption 2: Checking homoscedasticity of residuals i.e. the variance of residuals.
#
# Answers the question whether the standard deviation/variance of the residuals (errors)
# (for all predicted values) is measured properly and can be trusted.
#
# Homoscedasticity is checked by generating a residuals vs fitted values scatter plot.
# If points are spreading out or in, variance for residuals is not constant across
# the fitted values and hence residuals are not homoscedastic. 
#
# If this assumption is violated (i.e. residuals are not homoscedastic) the standard 
# deviation of the (forecast) errors cannot be measured properly and confidence intervals 
# will be too wide or too narrow. This means that in such situation, these values  
# cannot be useful or trusted. To fix this and under certain situations a log transformation
# of the dependent variable may fix this.
# 
##########################################################################################

# Reset graphical parameters
par(DEFAULT_GRAPHICAL_PARAMS)

# Set new graphical parameters
par(mfrow=c(2,2))
plot(studentPerfModel)


##########################################################################################
#
# Assumption 3: Checking normality of residuals 
#
# Answers the question whether the residuals have a normal distribution. 
# Residuals must be normaly distributed as a result of Central-Limit Theorem
# (see https://en.wikipedia.org/wiki/Central_limit_theorem)
#
# Normality of residuals can be checked visually by generating a Normal Q-Q plot of the
# residuals (see https://en.wikipedia.org/wiki/Q%E2%80%93Q_plot). If points significantly 
# diverge from the straight line - the “theoretical”, by default normal, quantile-quantile  
# plot - then normality of residuals is not supported.
#
# If this assumption is violated (i.e. residuals are not normaly distributed) this means 
# that there are problems determining if coefficients are significantly different from zero.
# To fix this, a non-linear transformation (e.g. log, power etc) of variables may address 
# the issue.
# 
##########################################################################################

# Reset graphical parameters since the plot
# requires different settings to appear properly
par(DEFAULT_GRAPHICAL_PARAMS)

# Drawing the Q-Q plot of residuals
qqnorm(studentPerfModel$residuals)
# Adds a line to a “theoretical”, by default normal, quantile-quantile plot
# If points lie on that line, residuals can be considered normally distributed
qqline(studentPerfModel$residuals, lty=5, lwd=1.3, col = 'red')





##########################################################################################
#
# Assumption 4: Checking for multicolinearity of independent variables in the case
#               of multiple linear regression. 
#
# Answers the question whether the independent variables of the linear model influence 
# each other. 
#
# If this assumption is violated (i.e. independent variables influence each other)  
# coefficients do not represent accurately the effects of variables on the dependent and
# p-values are larger for correlated independent variables hence hiding the impact of
# variables.
#
# Can be fixed by keeping one of the correlated variables.
# 
# 
##########################################################################################

par(DEFAULT_GRAPHICAL_PARAMS)


# Calculate and show variance inflation factor. This will
# give a quantitative measure of the colinearity between
# failures and absences
print(vif(studentPerfModel))

# Show the corrleation matrix. Will also show the level
# of multicolinearity between failures and absences. 

corrMatrix <- cor(studentMathematicsPerformanceData[, c('failures', 'absences')])
# Show the correlation matrix as a heatamap with color indicating
# the level of collinearity. 
# See https://cran.r-project.org/web/packages/corrplot/vignettes/corrplot-intro.html
# for how to interpret the plot.
corrplot(corrMatrix, method="color")

